/*
	STARMTCH.CPP
		-- a wildcard string match class 'symbmatchset'
		Developed using Turbo C++ version 3.0
		(if this won't compile on an older C++ compiler,
		 try moving the embedded structure symbmatch
		 out of symbmatchset)

		See the commented out 'main()'s  at the end of file
		for examples  of use.

		This is part of the POETICA project, a cooperative
		library of sound/text composition tools to be
		incorporated into the POETICA program.
		Improvements and/or additions to this code will be greatly
		appreciated. (Note, for example, that the matches don't
		support ignoring case)
				David Sharp  8/1/92
				GEnie     ---DSHARP3
				CompuServe---76646,1371
*/

#include <string.h>
#include <alloc.h>

const char CH = 'c';
const char FN = 'f';

class symbmatchset // a set of symbols and their respective match criteria
{
  protected:
	struct symbmatch // a symbol and its match criterion
	// embedded and protected so that symbmatchset knows
	// about all instances of symbmatch.
	{
		char   sym;
		char   matchtype; // 'c' for char_set, 'f' for matchfunc
		union{
			char * char_range;    // either a character set
			int (*matchfunc)(int);// or a 'boolean' fn
		};
		symbmatch(char * charset, char symbol)
			{ char_range=charset;  sym=symbol; matchtype=CH;}
		symbmatch(int (*func)(int), char symbol)
			{ matchfunc=func;      sym=symbol; matchtype=FN;}
		int matchp(char  ch);  //  is ch in char_range or does it
				      //  yield true fromm the matchfunc
	};
	//---------------------------------------------------------------
	int           numsyms;  // number of 'wild' symbols
	symbmatch * * matchset; // an array of symbol--match criteria pairs

  public:
	symbmatchset()
		{numsyms=0; matchset=NULL;}
	~symbmatchset();
	int    addsymbol(symbmatch * newsymbmatch);
	int    addsymbol(char * charset, char symbol);
	int    addsymbol(int (*matchfunc)(int), char symbol);
	symbmatch * findsym(char symbol); // returns the symbmatch
					  // for which ch is the symbol
	int    nmatchp(char * s, char * test, int n);
	char * search(char * tosearch, char * searchfor);
	int    var_nmatchp(char * s, char * test, int n);
	char * var_search(char * tosearch, char * searchfor);
};
int symbmatchset::symbmatch::matchp(char  ch)
{
	if(matchtype==CH){  // if char_range exists, use it to test for match
		if(strchr(char_range,ch))
			return 1;
	}
	else{ if(matchtype==FN) // otherwise if matchfunc exists, use it
		if((*matchfunc)(ch))
			return 1;
	}
	return 0;  // did not match (or neither method exists)
}
symbmatchset::~symbmatchset() // deallocate matchset and its constituents
{
	for(int n=0;n<numsyms;n++){
		if(matchset[n])
			delete matchset[n];
	}
	free(matchset);
}
int symbmatchset::addsymbol(symbmatch * newsymbmatch)
{
	if(numsyms)
		matchset=(symbmatch * *)
			  realloc(matchset,(numsyms+1)*(sizeof(symbmatch*)));
	else    matchset=(symbmatch * *)
			  malloc(sizeof(symbmatch*));
	if (matchset){
		matchset[numsyms]=newsymbmatch;
		numsyms++;
		return 1;
	}
	else
		return 0; //allocation failure
}

int symbmatchset::addsymbol(char * charset, char symbol)
{
	symbmatch * newsymbmatch = new symbmatch(charset,symbol);
	if(!newsymbmatch)
		return 0;
	if(addsymbol(newsymbmatch))
		return 1;
	else{
		delete newsymbmatch;
		return 0; //allocation failure
	}
}
int symbmatchset::addsymbol(int(*func)(int), char symbol)
{
	symbmatch * newsymbmatch = new symbmatch(func,symbol);
	if(!newsymbmatch)
		return 0;
	if(addsymbol(newsymbmatch))
		return 1;
	else{
		delete newsymbmatch;
		return 0; //allocation failure
	}
}
symbmatchset::symbmatch * symbmatchset::findsym(char ch)
{
	for (int n=0;n<numsyms;n++) {
		if (matchset[n]->sym==ch)
			return matchset[n];
	}
	return NULL; //not found
}
int symbmatchset::nmatchp(char * s, char * test, int n)
// see if first n chars of test match the first n chars of s,
{
	for(int i=0; i<n; i++){
		symbmatch * x;
		if( (x=findsym(test[i])) != NULL ){ // what symbol is it?
			if(!x->matchp(s[i])) // is s[i] in symbol's range?
				return 0;
		}
		else if(s[i] != test[i]) // both explicit, compare directly
				return 0;
	}
	return 1; //no clash found
}
int symbmatchset::var_nmatchp(char * s, char * test, int n)
// treats symbols in test as variables which are assigned a value
// when a match is found
{
	int testlen=strlen(test);
	char * temp = new char[testlen +1];
	strcpy(temp,test);// make a copy of test so we can mangle it
	for(int i=0; i<n; i++){
		symbmatch * x;
		if( (x=findsym(temp[i])) != NULL ){ // what symbol is it?
			if(!x->matchp(s[i])){ // is s[i] in symbol's range?
				delete temp;
				return 0;
			}
			else{ // replace all instances of the symbol found in
			      // test with the char s[i] which matched
				for(int j=i;j<testlen;j++){
					if(temp[j] == x->sym)
						temp[j]=s[i];
				}
			}
		}
		else if(s[i] != temp[i]){ // both explicit, compare directly
				delete temp;
				return 0;
		}
	}
	delete temp;
	return 1; //no clash found
}
char * symbmatchset::search(char * tosearch, char * searchfor)
// find first occurrence of 'searchfor' in 'tosearch'
{
	while((*tosearch) != '\0'){
		if(nmatchp(tosearch,searchfor,strlen(searchfor)))
			return tosearch;
		tosearch++;
	}
	return NULL;
}
char * symbmatchset::var_search(char * tosearch, char * searchfor)
// find first occurrence of 'searchfor' in 'tosearch' using var_nmatchp
{
	while((*tosearch) != '\0'){
		if(var_nmatchp(tosearch,searchfor,strlen(searchfor)))
			return tosearch;
		tosearch++;
	}
	return NULL;
}
//      --------------------- test examples ---------------------------------

/*
// ------------------------ example 1 --------------------------------
#include <iostream.h>

// for some test functions:
#include <ctype.h>

void main()
{
    char * s =    "bumblebee mom";
    for(int i=0;i<2;i++){ // go thru twice to test ~symbmatchset()
	symbmatchset wildcards;
	char vowels[6];
	if(i==0)
		strcpy(vowels,"aeiou");// 1st time thru should match
	else
		strcpy(vowels,"aiou"); // 2nd time shouldn't match
	if (  wildcards.addsymbol(vowels,'*') // '*' will match any vowel
	   && wildcards.addsymbol(isgraph,'#') // '#" matches any 'ink' char
	   && wildcards.addsymbol(isprint,'&')// '&' matches any print char
	   )
	{// addsymbol involves DOS memory reallocation

		char * test = "b*mbl#&**&m"; // a test string containing
					     // wildcard symbols
		cout<<"\n"<< i; // tell which run thru this is
		cout << "\n\""<<test<<"\"";
		if (wildcards.nmatchp(s,test,strlen(test)))  // do test
			cout <<"\n matches the first ";
		else cout <<"\n doesn't match the first " ;
		cout << strlen(test)<< " characters of\n"<< "\""<<s<<"\"";
		cout << "\n"<< coreleft();// see if ~symbmatchset() works:
					  // should be equal both times thru
	}
    }
}
*/
/*
//---------------------- example 2 ----------------------------------------
#include <iostream.h>

// for some test functions:
#include <ctype.h>

void main()
{
	char * text ="are you really the honest philosopher you say you are?";
	symbmatchset wildcards;

	wildcards.addsymbol(isgraph,'#'); // hard characters
	wildcards.addsymbol("aeiou",'*'); //vowels
	char * searchfor="#h*"; // hard_char--'h'--vowel. e.g. 'phi'
	char * here = text;
	while(here){
		if ( (here=wildcards.search(here,searchfor))!=NULL ){
			cout<<"\n"<<here;
			here++;
		}
	}
}
*/
/*
//---------------------- example 3 ----------------------------------------
#include <iostream.h>
*/
/*
int isconsonant(int ch) //define our own matching function
			// must be of type "int func(int)"
{
	char * vowels ="aeiouy"; // "... and sometimes y"
	if (strchr(vowels,ch))
		return 0;
	else return 1;
}
*/
/*

void main()
{
	char * text ="you have bloodied my keyboard, hercules.";
	symbmatchset wildcards;

	wildcards.addsymbol(isconsonant,'#'); // # = any consonant
	char * searchfor="#o";
	char * here = text;
	while(here){  // find all 'o's which are preceded by a consonant:
		if ( (here=wildcards.search(here,searchfor))!=NULL ){
			cout<<"\n"<<here;
			here++;
		}
	}
}
*/
//------------------------ example 4 -------------------------------------
// using var_search:
// ------------------ needs isconsonant() above -----------------------
/*
#include <iostream.h>
void main()
{
	char * text ="thats beet juice in the old beach boomer's boots, "
		     "his career, and hat";
	symbmatchset wildcards;
	wildcards.addsymbol(isconsonant,'@');
	wildcards.addsymbol("aeiouy",'#'); // # = any vowel
	char * searchfor="@##@"; // var_matchp will convert the second '#'and
				// '@' into whatever matches with the first
	char * here = text;
	while(here){  // find all 'b's followed by double vowels:
		if ( (here=wildcards.var_search(here,searchfor))!=NULL){
			cout<<"\n"<<here;
			here++;
		}
	}
}
*/